# ============
# Script to generate token frequency dict suing n_gram skill fullname
# ============

import json
import collections

with open('./skillNer/data/skill_db_relax_20.json', 'r+') as f:
    skills_db = json.load(f)


def get_dist_new(array):
    words = []
    for val in array:
        vals = val.split(' ')
        for v in vals:

            words.append(v)

    a = words
    counter = collections.Counter(a)
    counter = dict(counter)
    return counter


n_grams = [skills_db[key]['high_surfce_forms']['full']
           for key in skills_db if skills_db[key]['skill_len'] > 1]
n_gram_dist = get_dist_new(n_grams)
# save
with open('./skillNer/data/token_dist.json', 'w', encoding='utf-8') as f:
    json.dump(n_gram_dist, f, ensure_ascii=False, indent=4)



# ============
# Script to generate skill db that support multiple surface forms for the matching procedure

# the script use original emsi skill db can be downloaded using fetch.py
# download input db and place it in path
# ============
import re
import collections
import json

with open('./skillNer/data/skills_processed.json', 'r+') as f:
    SKILL_DB = json.load(f)
with open('./skillNer/data/token_dist.json', 'r+') as f:
    dist = json.load(f)
RELAX_PARAM = 0.2
new_skill_db = {}
for key in SKILL_DB:
    high_surface_form = {}
    low_surface_form = []
    match_on_tokens = False
    skill_len = SKILL_DB[key]['skill_len']
    skill_name = SKILL_DB[key]['skill_name']
    skill_type = SKILL_DB[key]['skill_type']
    clean_name = SKILL_DB[key]['skill_cleaned']
    skill_lemmed = SKILL_DB[key]['skill_lemmed']
    skill_stemmed = SKILL_DB[key]['skill_stemmed']
    abv = SKILL_DB[key]['abbreviation']
    # param diferentiate software and gramatical skills
    uni_match_on_stemmed = SKILL_DB[key]['match_on_stemmed']

    # high_surface_form.append(clean_name)
    # high_surface_form.append(skill_lemmed)
    if abv != '':

        high_surface_form['abv'] = abv
    # unigram skills
    if skill_len == 1:
        high_surface_form['full'] = clean_name
        if uni_match_on_stemmed:
            low_surface_form.append(skill_stemmed)

    if skill_len == 2:
        high_surface_form['full'] = skill_lemmed

        stemmed_tokens = skill_stemmed.split(' ')
        inv_stemmed_tokens = stemmed_tokens[::-1]
        low_surface_form.append(skill_stemmed)
        low_surface_form.append(' '.join(inv_stemmed_tokens))
        last = stemmed_tokens[-1]
        start = stemmed_tokens[0]
        if dist[last] == 1:
            low_surface_form.append(last)
        if dist[start]/dist[last] < RELAX_PARAM:
            low_surface_form.append(start)

    if skill_len > 2:
        high_surface_form['full'] = skill_lemmed
        match_on_tokens = True
    # write skill
    new_skill_db[key] = {'skill_name': skill_name,
                         'skill_type': skill_type,
                         'skill_len': skill_len,
                         'high_surfce_forms': high_surface_form,
                         'low_surface_forms': low_surface_form,
                         'match_on_tokens': match_on_tokens
                         }

list_ = []
for key in new_skill_db:
    low = new_skill_db[key]['low_surface_forms']
    skill_len = new_skill_db[key]['skill_len']
    if skill_len == 2:
        unique = [l for l in low if len(l.split(' ')) == 1]
        for a in unique:
            list_.append(a)


counter = collections.Counter(list_)

for key in new_skill_db:
    new_skill_db[key]['low_surface_forms']
    skill_len = new_skill_db[key]['skill_len']
    if skill_len == 2:
        low = new_skill_db[key]['low_surface_forms']
        new_l = []
        for l in low:
            if len(l.split(' ')) == 1:
                if counter[l] == 1:
                    new_l.append(l)
            else:
                new_l.append(l)
        new_skill_db[key]['low_surface_forms'] = new_l


# add n_gram skills surface forms
n_grams = [SKILL_DB[key] for key in SKILL_DB if SKILL_DB[key]['skill_len'] > 1]
rx = r"\b[A-Z](?=([&.]?))(?:\1[A-Z])+\b"


def extract_sub_forms(skill_name):
    return [x.group() for x in re.finditer(rx, skill_name)]


def remove_btwn_par(str_):
    return re.sub("[\(\[].*?[\)\]]", "", str_)


subs = []
for n_skill in n_grams:
    skill_name = n_skill['skill_name']
    new_skill_name = remove_btwn_par(skill_name)
    sub_f = extract_sub_forms(new_skill_name)
    if sub_f != []:
        print(skill_name)
        print(sub_f)
        print('--------')
        for s in sub_f:
            subs.append(s)
dist = collections.Counter(subs)
for key in new_skill_db:
    if new_skill_db[key]['skill_len'] > 2:

        skill_name = new_skill_db[key]['skill_name']
        new_skill_name = remove_btwn_par(skill_name)
        skill_low = new_skill_db[key]['low_surface_forms']
        sub_abv = extract_sub_forms(new_skill_name)
        for abv in sub_abv:
            if dist[abv] == 1:
                skill_low.append(abv)
        new_skill_db[key]['low_surface_forms'] = skill_low

# save
with open('./skillNer/data/skill_db_relax_20.json', 'w', encoding='utf-8') as f:
    json.dump(new_skill_db, f, ensure_ascii=False, indent=4)



# native packs
from typing import List
# installed packs
#
# my packs
from skillNer.cleaner import Cleaner, stem_text, find_index_phrase
from skillNer.general_params import S_GRAM_REDUNDANT


# building block of text
class Word:
    """Main data structure to hold metadata of words
    """

    def __init__(
        self,
        word: str
    ) -> None:
        """Construct an instance of Word

        Parameters
        ----------
        word : str
            The word is given as string

        Examples
        --------
        >>> from skillNer.text_class import Word
        >>> word_obj = Word("Hello")
        """

        # immutable version of word
        self.word = word

        # attributes
        self.lemmed = ""
        self.stemmed = ""

        self.is_stop_word = None
        self.is_matchable = True

        # position in sentence
        self.start: int
        self.end: int
        pass

    # get metadata of word
    def metadata(self) -> dict:
        """To get all metadata of the instance

        Returns
        -------
        dict
            dictionnary containing all metadata of object. Look at the example to see the returned keys

        Examples
        --------
        >>> from skillNer.text_class import Word
        >>> word_obj = Word("Hello")
        >>> word_obj.metadata().keys()
        dict_keys(['lemmed', 'stemmed', 'is_stop_word', 'is_matachable'])
        """

        return {
            "lemmed": self.lemmed,
            "stemmed": self.stemmed,
            "is_stop_word": self.is_stop_word,
            "is_matachable": self.is_matchable
        }

    # give the raw version of word when transformed to str
    def __str__(self) -> str:
        """To get the raw form of word

        Returns
        -------
        str
            raw form of word

        Examples
        --------
        >>> from skillNer.text_class import Word
        >>> word_obj = Word("Hello")
        >>> print(word_obj)
        Hello
        """
        return self.word

    # give the len of the word
    def __len__(self) -> int:
        """Gives the number of characters in word

        Returns
        -------
        int
            returns the number of characters in word

        Examples
        --------
        >>> from skillNer.text_class import Word
        >>> word_obj = Word("Hello")
        >>> len(word_obj)
        5
        """
        return len(self.word)


class Text:
    """The main object to store/preprocess a raw text. 
    The object behaviour is like a list according to words.
    """

    def __init__(
        self,
        text: str,
        nlp
    ):
        """Constructor of the class

        Parameters
        ----------
        text : str
            The raw text. It might be for instance a job description.
        nlp : [type]
            An NLP object instanciated from Spacy.

        Examples
        --------
        >>> import spacy
        >>> nlp = spacy.load('en_core_web_sm')
        >>> from skillNer.text_class import Text
        >>> text_obj = Text("Fluency in both English and French is mandatory")
        """

        # immutable version of text
        self.immutable_text = text

        # transformed text: lower + punctuation + extra space
        # this is the version of text that we will be working with
        cleaner = Cleaner(
            include_cleaning_functions=[
                "remove_punctuation",
                "remove_extra_space"
            ],
            to_lowercase=False
        )

        self.transformed_text = cleaner(text).lower()
        # abv version
        self.abv_text = cleaner(text)

        # list that holds all words within text
        self.list_words = []

        # construct list of words and create meta data object
        doc = nlp(self.transformed_text)

        for token in doc:
            # create word object
            word = Word(token.text)

            # lem and stem
            word.lemmed = token.lemma_
            word.stemmed = stem_text(token.text)

            # stop word and machability
            word.is_stop_word = token.is_stop
            # a stop word is unmatchable
            if token.is_stop:
                word.is_matchable = False

            self.list_words.append(word)

        # detect unmatchable words
        for redundant_word in S_GRAM_REDUNDANT:
            list_index = find_index_phrase(
                phrase=redundant_word, text=self.transformed_text)

            for index in list_index:
                self[index].is_matchable = False

    # return stemmed form of text either as str or list of words
    def stemmed(
        self,
        as_list: bool = False
    ):
        """To get the stemmed version of text

        Parameters
        ----------
        as_list : bool (default False)
            True to get a list of stemmed words within text. False, to get stemmed text in a form of string.

        Returns
        -------
        str | List[str]
            return the stemmed text in the specified form by the argument `as_list`.

        Examples
        --------
        >>> import spacy
        >>> nlp = spacy.load('en_core_web_sm')
        >>> from skillNer.text_class import Text
        >>> text_obj = Text("Fluency in both English and French is mandatory")
        >>> text_obj.stemmed()
        'fluenci in both english and french is mandatori'
        >>> text_obj.stemmed(as_list=True)
        ['fluenci', 'in', 'both', 'english', 'and', 'french', 'is', 'mandatori']
        """

        list_stems = [word.stemmed for word in self.list_words]

        if as_list:
            return list_stems

        return " ".join(list_stems)

    # return lemmed form of text either as str or list of words
    def lemmed(
        self,
        as_list: bool = False
    ):
        """To get the lemmed version of text

        Parameters
        ----------
        as_list : bool
            True to get a list of lemmed words within text. False, to get lemmed text in a form of string.

        Returns
        -------
        str | List[str]
            return the lemmed text in the specified form by the argument `as_list`

        Examples
        --------
        >>> import spacy
        >>> nlp = spacy.load('en_core_web_sm')
        >>> from skillNer.text_class import Text
        >>> text_obj = Text("Fluency in both English and French is mandatory")
        >>> text_obj.lemmed()
        'fluency in both english and french be mandatory'
        >>> text_obj.lemmed(as_list=True)
        ['fluency', 'in', 'both', 'english', 'and', 'french', 'be', 'mandatory']
        """

        list_lems = [word.lemmed for word in self.list_words]

        if as_list:
            return list_lems

        return " ".join(list_lems)

    # return raw version of text when converted to str
    def __str__(self) -> str:
        """To get the raw version of text

        Returns
        -------
        str
            returns the raw version of text

        Examples
        --------
        >>> import spacy
        >>> nlp = spacy.load('en_core_web_sm')
        >>> from skillNer.text_class import Text
        >>> text_obj = Text("Fluency in both English and French is mandatory")
        >>> print(text_obj)
        Fluency in both English and French is mandatory
        """

        return self.immutable_text

    # equip text with the behavior of a list
    # get item with []
    def __getitem__(
        self,
        index: int
    ) -> Word:
        """To get the word at the specified position by index

        Parameters
        ----------
        index : int
            the position of the word

        Returns
        -------
        Word
            returns thhe word object in the index-position

        Examples
        --------
        >>> import spacy
        >>> nlp = spacy.load('en_core_web_sm')
        >>> from skillNer.text_class import Text
        >>> text_obj = Text("Fluency in both English and French is mandatory")
        >>> text_obj[3]
        <skillNer.text_class.Word at 0x1cf13a9bd60>
        >>> print(text_obj[3])
        english
        """
        return self.list_words[index]

    # len of a text is the number of words in it
    def __len__(self) -> int:
        """To get the number of words in text

        Returns
        -------
        int
            returns the number of words in text

        Examples
        --------
        >>> import spacy
        >>> nlp = spacy.load('en_core_web_sm')
        >>> from skillNer.text_class import Text
        >>> text_obj = Text("Fluency in both English and French is mandatory")
        >>> len(text_obj)
        8
        """

        return len(self.list_words)

    # result a list of word object
    # each word contain the info of its start/end position
    @staticmethod
    def words_start_end_position(text: str) -> List[Word]:
        """To get the starting and ending index of each word in text

        Parameters
        ----------
        text : str
            The input text

        Returns
        -------
        List[Word]
            Returns a list of words where in each word the `start` and `end` 
            properties were filled by the starting and ending position of the word.

        Examples
        --------
        >>> import spacy
        >>> nlp = spacy.load('en_core_web_sm')
        >>> from skillNer.text_class import Text
        >>> list_words = Text.words_start_end_position("Hello World I am SkillNer")
        >>> word_1 = list_words[0]
        >>> print(word_1.start, word_1.end)
        0 5
        """
        # words in text
        list_words = []

        pointer = 0
        for raw_word in text.split(" "):
            # init object word
            word = Word(raw_word)

            # start and end of word
            word.start = pointer
            word.end = pointer + len(word)

            # update pointer
            pointer += len(word) + 1

            list_words.append(word)

        return list_words



# native packs
from typing import List


class Phrase:
    """Data structure to build html visualization of the annotated text.
    The input text is split into skill phrase and non skill phrase
    """

    def __init__(
        self,
        text: str
    ) -> None:

        # save raw text
        self.raw_text = text

        # position of phrase in original text
        self.start: int = None
        self.end: int = None

        # type of phrase: skill or not
        self.is_skill: bool = False

        # meta data
        self.skill_id: str = ""
        self.skill_name: str = ""
        self.skill_type: str = ""  # certification, hard skill or soft skill
        self.score: float = None
        self.type_matching: str = ""  # full match, n_gram

    def get_meta_data(self) -> dict:

        return {
            "skill name": self.skill_name,
            "matching type": self.type_matching,
            "score": self.score
        }

    @staticmethod
    def split_text_to_phare(
        annotation: dict,  # result of skill extractor
        SKILL_DB: dict
    ) -> List:
        """Main function to distinguish skill and non skill phrases

        Parameters
        ----------
        annotation : dict
            The output of the ``SkillExtractor.annotate``
        SKILL_DB: dict
            Data base of skills

        Returns
        -------
        List
            returns a list of phrases
        """
        # params
        list_words = annotation["text"].split(" ")

        # find skill phrases
        arr_skill_phrases = []

        for type_matching, arr_skills in annotation["results"].items():
            for skill in arr_skills:
                # create a phrase object
                phrase = Phrase(text=skill["doc_node_value"])
                phrase.is_skill = True

                # index word start and end
                start = skill["doc_node_id"][0]
                end = skill["doc_node_id"][-1]

                phrase.start = start
                phrase.end = end

                # meta data
                phrase.type_matching = type_matching
                phrase.skill_id = skill["skill_id"]
                phrase.skill_name = SKILL_DB[skill["skill_id"]]["skill_name"]
                phrase.skill_type = SKILL_DB[skill["skill_id"]]["skill_type"]
                phrase.score = skill["score"]

                # append phrase
                arr_skill_phrases.append(phrase)

        # handle case where no skill was annotated in text
        if not len(arr_skill_phrases):
            phrase = Phrase(text=annotation["text"])

            return [phrase]

        # order phrases
        arr_skill_phrases.sort(key=lambda item: item.start)

        # find non skill phrases
        arr_non_skill_phrases = []

        # handle case of first skill phrase
        start = 0
        end = arr_skill_phrases[0].start - 1

        non_skill_phrase = Phrase(" ".join(list_words[start:end + 1]))
        non_skill_phrase.start = start
        non_skill_phrase.end = end

        arr_non_skill_phrases.append(non_skill_phrase)

        # between skills phrases
        for i in range(len(arr_skill_phrases) - 1):
            start = arr_skill_phrases[i].end + 1
            end = arr_skill_phrases[i + 1].start - 1

            non_skill_phrase = Phrase(" ".join(list_words[start:end + 1]))
            non_skill_phrase.start = start
            non_skill_phrase.end = end

            arr_non_skill_phrases.append(non_skill_phrase)

        # handle case of last skill phrase
        # handle case of first skill phrase
        start = arr_skill_phrases[-1].end + 1
        end = len(list_words) - 1

        non_skill_phrase = Phrase(" ".join(list_words[start:end + 1]))
        non_skill_phrase.start = start
        non_skill_phrase.end = end

        arr_non_skill_phrases.append(non_skill_phrase)

        # merge arr non skill and skill then sort
        arr_phrases = arr_skill_phrases + arr_non_skill_phrases
        arr_phrases.sort(key=lambda item: item.start)

        return arr_phrases



# native packs
from typing import List
# installed packs
#
# my packs
from skillNer.text_class import Text


class Matchers:
    """class to instanciate a matcher pipeline used to annotate text.
    """

    def __init__(
        self,
        nlp,
        skills_db: dict,
        phraseMatcher
    ):
        """Constructor of the class

        Parameters
        ----------
        nlp : [type]
            NLP object loaded from spacy
        skills_db : dict
            A skill database that serves as a lookup table to annotate text
        phraseMatcher : [type]
            a phraseMatcher loaded using spacy
        """

        # params
        self.nlp = nlp
        self.skills_db = skills_db
        self.phraseMatcher = phraseMatcher
        #self.stop_words = stop_words

        # save matchers in a dict
        self.dict_matcher = {
            'full_matcher': self.get_full_matcher,
            'abv_matcher': self.get_abv_matcher,
            'full_uni_matcher': self.get_full_uni_matcher,
            'low_form_matcher': self.get_low_form_matcher,
            'token_matcher': self.get_token_matcher,
        }

        return

    # load specified matchers
    def load_matchers(
            self,
            include: List[str] = ['full_matcher',
                                  'abv_matcher',
                                  'full_uni_matcher',
                                  'low_form_matcher',
                                  'token_matcher',
                                  ],
            exclude: List[str] = []) -> dict:
        """To load matchers. The order of matchers define a pipeline.

        Parameters
        ----------
        include : List[str], optional
            List of matchers to include in the pipeline, by default ['full_matcher', 'abv_matcher', 'full_uni_matcher', 'low_form_matcher', 'token_matcher', ]
        exclude : List[str], optional
            List of matchers to exclude from the pipeline, by default []

        Returns
        -------
        dict
            returns a dictionnary where the keys are the name of matchers and the values are the matchers

        Examples
        --------
        >>> from skillNer.matcher_class import Matchers
        >>> from skillNer.general_params import SKILL_DB
        >>> import spacy
        >>> from spacy.matcher import PhraseMatcher
        >>> nlp = spacy.load('en_core_web_sm')
        >>> matcher_pipeline = Matchers(nlp, SKILL_DB, PhraseMatcher)
        >>> matcher_pipeline.load_matchers()
        loading full_matcher ...
        loading abv_matcher ...
        loading full_uni_matcher ...
        loading low_form_matcher ...
        loading token_matcher ...

        {'full_matcher': <spacy.matcher.phrasematcher.PhraseMatcher at 0x1c9680b8ac0>,
         'abv_matcher': <spacy.matcher.phrasematcher.PhraseMatcher at 0x1c96c90a6d0>,
         'full_uni_matcher': <spacy.matcher.phrasematcher.PhraseMatcher at 0x1c96c90a7b0>,
         'low_form_matcher': <spacy.matcher.phrasematcher.PhraseMatcher at 0x1c96c90a900>,
         'token_matcher': <spacy.matcher.phrasematcher.PhraseMatcher at 0x1c96c90a820>}
        """

        # #where to store loaded matchers
        loaded_matchers = {}

        # load matchers in if exclude is not empty
        # include will ignored
        if len(exclude):
            for matcher_name, matcher in self.dict_matcher.items():
                if matcher_name not in exclude:
                    print(f"loading {matcher_name} ...")
                    loaded_matchers[matcher_name] = matcher()
        else:
            for matcher_name, matcher in self.dict_matcher.items():
                if matcher_name in include:
                    print(f"loading {matcher_name} ...")
                    loaded_matchers[matcher_name] = matcher()

        return loaded_matchers

    # matchers
    # high confident matchers
    def get_full_matcher(self):
        # params
        nlp = self.nlp
        skills_db = self.skills_db
        full_matcher = self.phraseMatcher(nlp.vocab, attr="LOWER")

        # populate matcher
        for key in skills_db:
            # get skill info
            skill_id = key

            skill_len = skills_db[key]['skill_len']
            if skill_len > 1:
                skill_full_name = skills_db[key]['high_surfce_forms']['full']
                # add to matcher
                skill_full_name_spacy = nlp.make_doc(skill_full_name)
                full_matcher.add(str(skill_id), [skill_full_name_spacy])

        return full_matcher

    def get_abv_matcher(self):
        # params
        nlp = self.nlp
        skills_db = self.skills_db
        abv_matcher = self.phraseMatcher(nlp.vocab, attr="LOWER")

        # populate matcher
        for key in skills_db:
            # get skill info
            skill_id = key
            # check if there is a skill abrv
            if 'abv' in skills_db[key]['high_surfce_forms'].keys():
                skill_abv = skills_db[key]['high_surfce_forms']['abv']
                skill_abv_spacy = nlp.make_doc(skill_abv)
                abv_matcher.add(str(skill_id), [skill_abv_spacy])

        return abv_matcher

    def get_full_uni_matcher(self):
        # params
        nlp = self.nlp
        skills_db = self.skills_db
        full_uni_matcher = self.phraseMatcher(nlp.vocab, attr="LOWER")

        # populate matcher
        for key in skills_db:
            # get skill info
            skill_id = key

            skill_len = skills_db[key]['skill_len']
            if skill_len == 1:
                skill_full_name = skills_db[key]['high_surfce_forms']['full']
                # add to matcher
                skill_full_name_spacy = nlp.make_doc(skill_full_name)
                full_uni_matcher.add(str(skill_id), [skill_full_name_spacy])

        return full_uni_matcher

    # low confident matchers
    def get_low_form_matcher(self):
        # params
        nlp = self.nlp
        skills_db = self.skills_db
        low_form_matcher = self.phraseMatcher(nlp.vocab, attr="LOWER")

        # populate matcher
        for key in skills_db:

            # get skill info
            skill_id = key
            skill_len = skills_db[key]['skill_len']

            low_surface_forms = skills_db[key]['low_surface_forms']
            for form in low_surface_forms:
                skill_form_spacy = nlp.make_doc(form)
                low_form_matcher.add(str(skill_id), [skill_form_spacy])
        return low_form_matcher

    def get_token_matcher(self):
        # params
        nlp = self.nlp
        skills_db = self.skills_db
        token_matcher = self.phraseMatcher(nlp.vocab, attr="LOWER")

        # populate matcher
        for key in skills_db:
            # get skill info
            skill_id = key
            match_on_tokens = skills_db[key]['match_on_tokens']

            if match_on_tokens:  # check if skill accept matches on its unique tokens
                skill_lemmed = skills_db[key]['high_surfce_forms']['full']
                skill_lemmed_tokens = skill_lemmed.split(' ')

                # add tokens to matcher
                for token in skill_lemmed_tokens:
                    # give id that ref 1_gram matching
                    if token.isdigit():
                        pass
                    else:
                        id_ = skill_id
                        token_matcher.add(str(id_), [nlp.make_doc(token)])

        return token_matcher


class SkillsGetter:
    """Class that gather functions to get the matched skills.
    """

    def __init__(
        self,
        nlp
    ):

        # param
        self.nlp = nlp
        return

    def get_full_match_skills(
        self,
        text_obj: Text,
        matcher
    ):

        skills = []
        doc = self.nlp(text_obj.lemmed())

        for match_id, start, end in matcher(doc):
            id_ = matcher.vocab.strings[match_id]
            # add full_match to store
            skills.append({'skill_id': id_,
                           'doc_node_value': str(doc[start:end]),
                           'score': 1,
                           'doc_node_id': list(range(start, end))})
            # mutate text tokens metadata (unmatch attr)
            for token in text_obj[start:end]:
                token.is_matchable = False

        return skills, text_obj

    def get_abv_match_skills(
        self,
        text_obj: Text,
        matcher
    ):
        skills = []

        doc = self.nlp(text_obj.abv_text)
        for match_id, start, end in matcher(doc):
            id_ = matcher.vocab.strings[match_id]
            if text_obj[start].is_matchable:
                skills.append({'skill_id': id_,
                               'score': 1,
                               'doc_node_value': str(doc[start:end]),
                               'doc_node_id': [start]})
                # mutate matched tokens
                for token in text_obj[start:end]:
                    token.is_matchable = False

        return skills, text_obj

    def get_full_uni_match_skills(
        self,
        text_obj: Text,
        matcher
    ):

        skills = []

        doc = self.nlp(text_obj.transformed_text)
        for match_id, start, end in matcher(doc):
            id_ = matcher.vocab.strings[match_id]
            if text_obj[start].is_matchable:
                skills.append({'skill_id': id_+'_fullUni',
                               'score': 1,
                               'doc_node_value': str(doc[start:end]),
                               'doc_node_id': [start],
                               'type': 'full_uni'})

        return skills, text_obj

    def get_token_match_skills(
        self,
        text_obj: Text,
        matcher
    ):

        skills_full = []
        skills = []
        sub_matches = []
        full_matches = []

        doc = self.nlp(text_obj.lemmed())
        for match_id, start, end in matcher(doc):
            id_ = matcher.vocab.strings[match_id]

            # add
            if text_obj[start].is_matchable:
                skills.append({'skill_id': id_+'_oneToken',
                               'doc_node_value': str(doc[start:end]),
                               'doc_node_id': [start],
                               'type': 'one_token'})

        return skills

    def get_low_match_skills(
        self,
        text_obj: Text,
        matcher
    ):

        skills = []
        doc = self.nlp(text_obj.stemmed())

        for match_id, start, end in matcher(doc):
            id_ = matcher.vocab.strings[match_id]
            # handle skill in the end of phrase
            start = start if start < len(text_obj) else start - 1
            if text_obj[start].is_matchable:
                skills.append({'skill_id': id_+'_lowSurf',
                               'doc_node_value': str(doc[start:end]),
                               'doc_node_id': list(range(start, end)),
                               'type': 'lw_surf'})

        return skills, text_obj



# native packs
import os
import json
from posixpath import dirname
# installed packs
#
# my packs
from skillNer.network.remote_db import RemoteBucket

# mapping skill and color
SKILL_TO_COLOR = {
    'Hard Skill': '#818CF8',
    'Soft Skill': '#F472B6',
    'Certification': "#552448"
}

# mapping skill and color tail
SKILL_TO_COLOR_TAILWIND = {
    'Hard Skill': 'red-500',
    'Soft Skill': 'blue-500',
    'Certification': "green-500"
}


# init remote bucket to fetch db in case they don't exist locally
bucket = RemoteBucket(
    branch="first_release"
)

# load local data
# else fetch remote data and save it
try:
    with open('skill_db_relax_20.json') as json_file:
        SKILL_DB = json.load(json_file)
except:
    SKILL_DB = bucket.fetch_remote("SKILL_DB")
    # dump data
    with open('skill_db_relax_20.json', 'w') as fp:
        json.dump(SKILL_DB, fp)

# load token distribution dict
try:
    with open('token_dist.json') as json_file:
        TOKEN_DIST = json.load(json_file)
except:
    TOKEN_DIST = bucket.fetch_remote("TOKEN_DIST")
    # dump data
    with open('token_dist.json', 'w') as fp:
        json.dump(TOKEN_DIST, fp)


# list of punctuation
LIST_PUNCTUATIONS = ['/', '·', ',', '.',
                     '-', '(', ')', ',', ':', '!', "'", '?']

# list of redundant words to remove from text
S_GRAM_REDUNDANT = [
    'no no',
    'of the',
    'you have',
    'in the',
    'experience in',
    'ability to',
    'knowledge of',
    'pdf format',
    'at least',
    'experience of',
    'you are',
    'mastery of',
    'to the',
    'least years',
    'years of',
    'sense of',
    'years in',
    'or equivalent',
    'of at',
    'professional experience',
    'the field',
    'in similar',
    'copy of',
    'field of',
    'of experience',
    'similar position',
    'on the',
    'ofppt ma',
    'level of',
    'in pdf',
    'to justify',
    'will be',
    'good knowledge',
    'to be',
    'of years',
    'you justify',
    'justify cumulative',
    'an asset',
    'degree in',
    'cumulative professional',
    'access to',
    'in computer',
    'to work',
    'have good',
    'how to',
    'for the',
    'related to',
    'to hold',
    'the position',
    'an experience',
    'able to',
    'very good',
    'mastering the',
    'diploma giving',
    'the public',
    'giving access',
    'public service',
    'training in',
    'must be',
    'diploma diploma',
    'you re',
    'and you',
    'required diploma',
    'is an',
    'by foreign',
    'foreign institutions',
    'institutions must',
    'accompanied by',
    'copies of',
    'of certificates',
    'certificates of',
    'of equivalence',
    'recto verso',
    'downloadable from',
    'www ofppt',
    'experience related',
    'the required',
    'be accompanied',
    'by copies',
    'application form',
    'the cni',
    'cni recto',
    'issued by',
    'equivalence pdf',
    'verso in',
    'format one',
    'one page',
    'page only',
    'diploma issued',
    'and or',
    'position pdf',
    'bac training',
    'the job',
    'computer science',
    'format downloadable',
    'updated cv',
    'from www',
    'detailing the',
    'team spirit',
    'for information',
    'cv detailing',
    'job experience',
    'technician or',
    'have an',
    'information pdf',
    'spirit of',
    'would be',
    'than that',
    'oral and',
    'and written',
    'and oral',
    'years experience',
    'good level',
    'of moroccan',
    'be considered',
    'or more',
    'the following',
    'the application',
    'of age',
    'the offer',
    'or similar',
    'of applications',
    'the date',
    'date of',
    'moroccan nationality',
    'minimum experience',
    'and apply',
    'an application',
    'form f1',
    'other than',
    'will not',
    'not be',
    'candidature file',
    'to submit',
    'submit an',
    'by sending',
    'deadline for',
    'for receipt',
    'receipt of',
    'applications is',
    '2021 at',
    'at midnight',
    'bac in',
    'with the',
    'transmitted via',
    'via channel',
    'channel other',
    'that mentioned',
    'mentioned above',
    'above will',
    'analysis and',
    'you to',
    'register and',
    'sending the',
    'application file',
    'file in',
    'the website',
    'minimum of',
    'experience as',
    'know how',
    'http recruitment',
    'recruitment ofppt',
    'format to',
    'website http',
    'perfect mastery',
    'bachelor degree',
    'age on',
    'f1 for',
    'have the',
    'any file',
    'higher education',
    'command of',
    'file transmitted',
    'technical skills',
    'good relationship',
    'entry into',
    'into service',
    'engineering school',
    'of entry',
    'an excellent',
    'this position',
    'knowledge in',
    'of service',
    'good mastery',
    'scale of',
    'july 2021',
    'and the',
    'of analysis',
    'you master',
    'equivalent in',
    'have at',
    'writing and',
    'want to',
    'as team',
    'what you',
    'first experience',
    'technical expertise',
    'training bac',
    'is that',
    'that what',
    're saying',
    'master degree',
    'to manage',
    'to years',
    'at the',
    'years as',
    'offer you',
    'is plus',
    'a good',
    'fluency in'
]



# native packs
import requests


MAPPING_NAME_URL = {
    "SKILL_DB": "buckets/skill_db_relax_20.json",
    "TOKEN_DIST": "buckets/token_dist.json"
}


class RemoteBucket:
    """Main class to fetch data bases (db) from repo. db are saved in a `.json` files
    """

    def __init__(
        self,
        token: str = "",
        branch: str = "master"
    ) -> None:
        """Constructor of the class

        Parameters
        ----------
        token : str, optional
            Your GitHub token in case repo is private, by default "" which is the case of public repo
        branch : str, optional
            the branch from which to fetch db, by default "master"
        """

        # save params
        self.token = token
        self.branch = branch

        # construct endpoint
        self.end_point = f"https://raw.githubusercontent.com/AnasAito/SkillNER/{self.branch}"
        return

    def fetch_remote(
        self,
        db_name: str
    ) -> dict:
        """Function to fetch db

        Parameters
        ----------
        db_name : str in ["SKILL_DB", "TOKEN_DIST"]
            Name of the db to fetch

        Returns
        -------
        dict
            returns the db in format of a python dict object

        Examples
        --------
        >>> from skillNer.network.remote_db import RemoteBucket
        >>> buckets = RemoteBucket(
            branch="master"
        )
        >>> buckets.fetch_remote("SKILL_DB")
        ...
        """
        # request props
        url = f"{self.end_point}/{MAPPING_NAME_URL[db_name]}"

        # check if repo is private
        if self.token:
            headers = {
                'Authorization': f'token {self.token}'
            }
        else:
            headers = {}

        # fetch
        response = requests.get(
            url=url,
            headers=headers
        )

        # return content in json format
        return response.json()



# native packs
#
# installed packs
# my packs
from skillNer.text_class import Text
from skillNer.matcher_class import Matchers, SkillsGetter
from skillNer.utils import Utils
from skillNer.general_params import SKILL_TO_COLOR

from skillNer.visualizer.phrase_class import Phrase


class SkillExtractor:
    """Main class to annotate skills in a given text and visualize them.
    """

    def __init__(
        self,
        nlp,
        skills_db,
        phraseMatcher,
        tranlsator_func=False
    ):
        """Constructor of the class.

        Parameters
        ----------
        nlp : [type]
            NLP object loaded from spacy.
        skills_db : [type]
            A skill database used as a lookup table to annotate skills.
        phraseMatcher : [type]
            A phrasematcher loaded from spacy.
        tranlsator_func :Callable
            A fucntion to translate text from source language to english def tranlsator_func(text_input: str) -> text_input:str
        """

        # params
        self.tranlsator_func = tranlsator_func
        self.nlp = nlp
        self.skills_db = skills_db
        self.phraseMatcher = phraseMatcher

        # load matchers: all
        self.matchers = Matchers(
            self.nlp,
            self.skills_db,
            self.phraseMatcher,
            # self.stop_words
        ).load_matchers()

        # init skill getters
        self.skill_getters = SkillsGetter(self.nlp)

        # init utils
        self.utils = Utils(self.nlp, self.skills_db)
        return

    def annotate(
        self,
        text: str,
        tresh: float = 0.5
    ) -> dict:
        """To annotate a given text and thereby extract skills from it.

        Parameters
        ----------
        text : str
            The target text.
        tresh : float, optional
            A treshold used to select skills in case of confusion, by default 0.5

        Returns
        -------
        dict
            returns a dictionnary with the text that was used and the annotated skills (see example).

        Examples
        --------
        >>> import spacy
        >>> from spacy.matcher import PhraseMatcher
        >>> from skillNer.skill_extractor_class import SkillExtractor
        >>> from skillNer.general_params import SKILL_DB
        >>> nlp = spacy.load('en_core_web_sm')
        >>> skill_extractor = SkillExtractor(nlp, SKILL_DB, PhraseMatcher)
        loading full_matcher ...
        loading abv_matcher ...
        loading full_uni_matcher ...
        loading low_form_matcher ...
        loading token_matcher ...
        >>> text = "Fluency in both english and french is mandatory"
        >>> skill_extractor.annotate(text)
        {'text': 'fluency in both english and french is mandatory',
        'results': {'full_matches': [],
        'ngram_scored': [{'skill_id': 'KS123K75YYK8VGH90NCS',
            'doc_node_id': [3],
            'doc_node_value': 'english',
            'type': 'lowSurf',
            'score': 1,
            'len': 1},
        {'skill_id': 'KS1243976G466GV63ZBY',
            'doc_node_id': [5],
            'doc_node_value': 'french',
            'type': 'lowSurf',
            'score': 1,
            'len': 1}]}}
        """

        # check translator
        if self.tranlsator_func:
            text = self.tranlsator_func(text)

        # create text object
        text_obj = Text(text, self.nlp)
        # get matches
        skills_full, text_obj = self.skill_getters.get_full_match_skills(
            text_obj, self.matchers['full_matcher'])

        # tests

        skills_abv, text_obj = self.skill_getters.get_abv_match_skills(
            text_obj, self.matchers['abv_matcher'])

        skills_uni_full, text_obj = self.skill_getters.get_full_uni_match_skills(
            text_obj, self.matchers['full_uni_matcher'])

        skills_low_form, text_obj = self.skill_getters.get_low_match_skills(
            text_obj, self.matchers['low_form_matcher'])

        skills_on_token = self.skill_getters.get_token_match_skills(
            text_obj, self.matchers['token_matcher'])
        full_sk = skills_full + skills_abv
        # process pseudo submatchers output conflicts
        to_process = skills_on_token + skills_low_form + skills_uni_full
        process_n_gram = self.utils.process_n_gram(to_process, text_obj)

        return {
            'text': text_obj.transformed_text,
            'results': {
                'full_matches': full_sk,
                'ngram_scored': [match for match in process_n_gram if match['score'] >= tresh],

            }
        }


import setuptools


dependencies = [
    "numpy",
    "pandas",
    "nltk",
    "spacy",
    "jellyfish",
    "scipy"
]

setuptools.setup(
    name="NUGigSkillNER",
    version="1.0.2",
    author="Ethan Mandel",
    author_email="emandel2630@gmail.com",
    description="An NLP module to automatically Extract skills and certifications from unstructured job postings, texts, and applicant's resumes",
    url="https://github.com/emandel2630/NUGigSkillNER",
    keywords=["skillNer", 'python', 'NLP', "NER",
              "skills-extraction", "job-description"],
    packages=setuptools.find_packages(),
    install_requires=dependencies,
    classifiers=[
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: MIT License',
        "Operating System :: OS Independent",
    ],
)



# process_n_gram, process_uni_gram : filter and score

# native packs
import collections
import functools
import math

# installed packs
import numpy as np
import jellyfish
# my packs
from skillNer.text_class import Text
from skillNer.general_params import TOKEN_DIST
from scipy.sparse import csr_matrix
import pandas as pd

class Utils:
    def __init__(self, nlp, skills_db):
        self.nlp = nlp
        self.skills_db = skills_db
        self.token_dist = TOKEN_DIST
        self.sign = functools.partial(math.copysign, 1)
        return

    def make_one(self, cluster, len_):
        a = [1] * len_
        return [1*(i in cluster) for i, one in enumerate(a)]

    def split_at_values(self, lst, val):
        return [i for i, x in enumerate(lst) if x != val]

    def grouper(self, iterable, dist):
        prev = None
        group = []
        for item in iterable:
            if prev == None or item - prev <= dist:
                group.append(item)
            else:
                yield group
                group = [item]
            prev = item
        if group:
            yield group

    def get_clusters(self, co_oc):
        clusters = []
        for i,row in enumerate(co_oc.tolil().rows):
            if len(row)== 0:
                continue
            # divide row into clusters deivided by 0 : 0 occurence where clusters refer to token id
            # example [2,3,0,5,0,0] -> [0,1,3] -> [[0,1],[3]]
            # clusts = list(self.grouper(self.split_at_values(row, 0), 1))
            clusts = list(self.grouper(row, 1))
                # select token relative cluster via its idx
               # example i==0 => [[0,1],[3]] => a = [0,1]
            a = [c for c in clusts if i in c][0]
            if a not in clusters:
                clusters.append(a)
                    
        # return unique clusters [token id]
        return clusters

    def get_corpus(self, text, matches):
        """create a corpus matrix which will be used in future computations.

           Parameters
           ----------
           matches (list): list of matches generated by sub matchers
           text (Text): text object

           Returns
           -------

               corpus : return binary matrix   => (n :skills matched )* (m : tokens in text )  
                                                1 : skill contains token
                                                0 : otherwise
               look_up : return a mapper from skill_ids to its equivalent row index in corpus
        """
        
        len_ = len(text)
        unique_skills = list(set([match['skill_id'] for match in matches]))
        skill_text_match_bin = [0]*len_
        match_df = pd.DataFrame(matches)
        match_df_group = match_df.groupby('skill_id')['doc_node_id']
        corpus=[]
        look_up = {}
        on_inds =[]
        for idx,gr in enumerate(match_df_group):
            skill_id =gr[0]
            g=gr[1]
            skill_text_match_bin = [0]*len_
            look_up[idx]=skill_id
            on_inds = [j for sub in g for j in sub]
            skill_text_match_bin_updated = [(i in on_inds)*1 for i, _ in enumerate(skill_text_match_bin)]

            corpus.append(skill_text_match_bin_updated)
        return np.array(corpus), look_up

    def one_gram_sim(self, text_str, skill_str):
        # transform into sentence
        text = text_str + ' ' + skill_str
        tokens = self.nlp(text)
        token1, token2 = tokens[0], tokens[1]
        try:
            vec_similarity = token1.similarity(token2)
            return vec_similarity
        except:
            # try Levenshtein Distance  if words not found in spacy corpus
            str_distance_similarity = jellyfish.jaro_distance(
                text_str.lower(), skill_str.lower())
            return str_distance_similarity

    def compute_w_ratio(self, skill_id, matched_tokens):
        skill_name = self.skills_db[skill_id]['high_surfce_forms']['full'].split(
            ' ')
        skill_len = self.skills_db[skill_id]['skill_len']
        # favorize the matched tokens uphead
        late_match_penalty_coef = 0.1
        token_ids = sum([(1-late_match_penalty_coef*skill_name.index(token))
                         for token in matched_tokens])

        return token_ids/skill_len

    def retain(self, text_obj, span, skill_id, sk_look, corpus):
        """ add doc here  """
        real_id, type_ = sk_look[skill_id].split('_')

        # get skill len
        len_ = self.skills_db[real_id]['skill_len']
        # get intersection length of full  skill name  and span tokens
        len_condition = corpus[skill_id].dot(span)

        # start :to be deleted
        s_gr = np.array(list(span))*np.array(list(corpus[skill_id]))
        def condition(x): return x == 1

        s_gr_n = [idx for idx, element in enumerate(
            s_gr) if condition(element)]
        # end

        if type_ == 'oneToken':
            # if skill is n_gram (n>2)
            score = self.compute_w_ratio(
                real_id, [text_obj[ind].lemmed for ind in s_gr_n])

        if type_ == 'fullUni':
            score = 1

        if type_ == 'lowSurf':
            if len_ > 1:

                score = sum(s_gr)

            else:
                # if skill is uni_gram (n=1)
                text_str = ' '.join([str(text_obj[i])
                                     for i, val in enumerate(s_gr) if val == 1])
                skill_str = self.skills_db[real_id]['high_surfce_forms']['full']

                score = self.one_gram_sim(text_str, skill_str)

        return {'skill_id': real_id,
                'doc_node_id':  [i for i, val in enumerate(s_gr) if val == 1],
                'doc_node_value': ' '.join([str(text_obj[i]) for i, val in enumerate(s_gr) if val == 1]),
                'type': type_,
                'score': score,
                'len': len_condition
                }
    # main functions

    def process_n_gram(self, matches, text_obj: Text):
        """apply on conflicted matches to choose which  ones to keep

           Parameters
           ----------
           matches (list): list of matches generated by sub matchers
           text_obj (Text): text object 

           Returns
           -------

               list: return choosen skills with their given words span in the text and thir score  

           """
        if len(matches) == 0:
            return matches

        text_tokens = text_obj.lemmed(as_list=True)
        len_ = len(text_tokens)

        corpus, look_up = self.get_corpus(text_tokens, matches)
        corpus_csr = csr_matrix(corpus)
        # generate spans (a span is a list of tokens where one or more skills are matched)

        # co-occurence of tokens aij : co-occurence count of token i with token j
        # co_occ = np.matmul(corpus.T, corpus)
        co_occ_csr = corpus_csr.T.dot(corpus_csr)
        # create spans of tokens that co-occured
        clusters = self.get_clusters(co_occ_csr)

        # one hot encoding of clusters
        # example [0,1,2] => [1,1,1,0,0,0] , encoding length  = text length
        ones = [self.make_one(cluster, len_) for cluster in clusters]
        # generate list of span and list of skills that have conflict on spans [(span,[skill_id])]
        spans_conflicts = [(np.array(one), np.array([a_[0] for a_ in np.argwhere(corpus_csr.dot(one) != 0)]))
                           for one in ones]

        # filter and score
        new_spans = []
        for span_conflict in spans_conflicts:
            span, skill_ids = span_conflict
            span_scored_skills = []
            types = []
            scores = []
            lens = []
            for sk_id in skill_ids:
                # score skill given span
                scored_sk_obj = self.retain(
                    text_obj, span, sk_id, look_up, corpus)
                span_scored_skills.append(scored_sk_obj)
                types.append(scored_sk_obj['type'])
                lens.append(scored_sk_obj['len'])
                scores.append(scored_sk_obj['score'])
            # extract best candiate for a given span
            if 'oneToken' in types and len(set(types)) > 1:
                # having a ngram skill with other types in span condiates :
                # priotize skills with high match length if length >1
                id_ = np.array(scores).argmax()
                max_score = 0.5  # selection treshold
                for i, len_ in enumerate(lens):
                    if len_ > 1 and types[i] == 'oneToken':
                        if scores[i] >= max_score:
                            id_ = i

                new_spans.append(span_scored_skills[id_])

            else:
                max_score_index = np.array(scores).argmax()
                new_spans.append(span_scored_skills[max_score_index])

        return new_spans



# installed packs
from nltk.stem import PorterStemmer
# import en_core_web_lg
# native packs
from typing import List
# my pack
from skillNer.general_params import S_GRAM_REDUNDANT, LIST_PUNCTUATIONS


# load nlp
# nlp = en_core_web_lg.load()
# list of cleaning functions names
all_cleaning = [
    "remove_punctuation",
    "remove_redundant",
    "stem_text",
    "lem_text",
    "remove_extra_space"
]


# remove punctuation from text
def remove_punctuation(
    text: str,
    list_punctuations: List[str] = LIST_PUNCTUATIONS,
) -> str:
    """To Remove punctuation from a given text.

    Parameters
    ----------
    text : str
        The text to clean.
    list_punctuations: List[str], optional
        A list that define the punctuations to remove, by default LIST_PUNCTUATIONS

    Returns
    -------
    str
        returns a the provided text after removing all punctuations

    Examples
    --------
    >>> from SkillNer.cleaner import remove_punctuation
    >>> text = "Hello there, I am SkillNer! Annoation, annotation, annotation ..."
    >>> print(remove_punctuation(text))
    Hello there  I am SkillNer  Annoation  annotation  annotation
    """

    for punc in list_punctuations:
        text = text.replace(punc, " ")

    # use .strip() to remove extra space in the begining/end of the text
    return text.strip()


# remove redundant words
def remove_redundant(
    text: str,
    list_redundant_words: List[str] = S_GRAM_REDUNDANT,
) -> str:
    """To remove phrases that appear frequently and that can not be used to infere skills.

    Parameters
    ----------
    text : str
        The text to clean.
    list_redundant_words : List[str], optional
        The list of phrases to remove, by default S_GRAM_REDUNDANT

    Returns
    -------
    str
        returns text after removing all redundant words provided in `list_redundant_words`

    Examples
    --------
    >>> from SkillNer.cleaner import remove_redundant
    >>> text = "you have professional experience building React apps, you are familiar with version control using git and GitHub"
    >>> print(remove_redundant(text))
    building React apps,  familiar with version control using git and GitHub
    """

    for phrase in list_redundant_words:
        text = text.replace(phrase, "")

    # use .strip() to remove extra space in the begining/end of the text
    return text.strip()


# stem using a predefined stemer
def stem_text(
    text: str,
    stemmer=PorterStemmer(),
) -> str:
    """To stem a text 

    Parameters
    ----------
    text : str
        The text to be stemmed.
    stemmer : stemmer loaded from nltk, optional
        The stemmer to be used when stemming text, by default PorterStemmer()

    Returns
    -------
    str
        returns text after stemming it.

    Examples
    --------
    >>> from SkillNer.cleaner import stem_text
    >>> text = "you have professional experience building React apps, you are familiar with version control using git and GitHub"
    >>> print(stem_text(text))
    you have profession experi build react apps, you are familiar with version control use git and github
    """

    return " ".join([stemmer.stem(word) for word in text.split(" ")])


# lem text using nlp loaded from scapy
def lem_text(
    text: str,
    nlp,
) -> str:
    """To lem a text.

    Parameters
    ----------
    text : str
        the text to be lemmed
    nlp : nlp object loaded form spacy.
        the nlp used to lem the text

    Returns
    -------
    str
        returns text after lemming it.

    Examples
    --------
    >>> from SkillNer.cleaner import lem_text
    >>> import spacy
    >>> nlp = spacy.load("en_core_web_sm")
    >>> text = "you have professional experience building React apps, you are familiar with version control using git and GitHub"
    >>> print(lem_text(text, nlp))
    you have professional experience building react app , you be familiar with version control use git and GitHub
    """

    doc = nlp(text)
    return ' '.join([token.lemma_ for token in doc])


# remove extra space
def remove_extra_space(
    text: str,
) -> str:
    """To remove extra space in a given text.

    Parameters
    ----------
    text : str
        The text to clean.

    Returns
    -------
    str
        returns text after removing all redundant spaces.

    Examples
    --------
    >>> from SkillNer.cleaner import remove_extra_space
    >>> text = " I am   sentence with   a lot of  annoying extra    spaces    ."
    >>> print(remove_extra_space(text))
    I am sentence with a lot of annoying extra spaces .
    """

    return " ".join(text.split())


# gather all cleaning functions in a dict
dict_cleaning_functions = dict_cleaning_functions = {
    "remove_punctuation": remove_punctuation,
    "remove_redundant": remove_redundant,
    "stem_text": stem_text,
    "lem_text": lem_text,
    "remove_extra_space": remove_extra_space
}


# find index of words of a phrase in a text
# return an empty list if phrase not in text
def find_index_phrase(
    phrase: str,
    text: str,
) -> List[int]:
    """Function to determine the indexes of words in a phrase given a text.

    Parameters
    ----------
    phrase : str
        the input phrase.
    text : str
        the text where to look for phrase and detemine their indexes

    Returns
    -------
    List[int]
        returns a list of the indexes of words in phrase. An empty list is returned if phrase is not in text.

    Examples
    --------
    >>> from SkillNer.cleaner import find_index_phrase
    >>> text = "you have professional experience building React apps, you are familiar with version control using git and GitHub"
    >>> phrase = "experience building"
    >>> find_index_phrase(phrase, text)
    [3, 4]
    >>> find_index_phrase(phrase="Hello World", text)
    []
    """

    if phrase in text:
        # words in text
        list_words = text.split(" ")

        # words in phrase
        list_phrase_words = phrase.split(" ")
        n = len(list_phrase_words)

        for i in range(len(text) - n):
            if list_words[i:i + n] == list_phrase_words:
                return [i + k for k in range(n)]

    return []


class Cleaner:
    """A class to build pipelines to clean text.
    """

    def __init__(
        self,
        to_lowercase: bool = True,
        include_cleaning_functions: List[str] = all_cleaning,
        exclude_cleaning_function: List[str] = [],
    ):
        """the constructor of the class.

        Parameters
        ----------
        to_lowercase : bool, optional
            whether to lowercase the text before cleaning it, by default True
        include_cleaning_functions : List, optional
            List of cleaning operations to include in the pipeline, by default all_cleaning
        exclude_cleaning_function : List, optional
            List of cleaning operations to exclude for the pipeline, by default []
        """

        # store params
        self.include_cleaning_functions = include_cleaning_functions
        self.exclude_cleaning_functions = exclude_cleaning_function
        self.to_lowercase = to_lowercase

    def __call__(
        self,
        text: str
    ) -> str:
        """To apply the initiallized cleaning pipeline on a given text.

        Parameters
        ----------
        text : str
            text to clean

        Returns
        -------
        str
            returns the text after applying all cleaning operations on it.

        Examples
        -------
        >>> from skillNer.cleaner import Cleaner
        >>> cleaner = Cleaner(
                        to_lowercase=True,
                        include_cleaning_functions=["remove_punctuation", "remove_extra_space"]
                    )
        >>> text = " I am   sentence with   a lot of  annoying extra    spaces    , and !! some ,., meaningless punctuation ?! .! AH AH AH"
        >>> cleaner(text)
        'i am sentence with a lot of annoying extra spaces and some meaningless punctuation ah ah ah'
        """

        # lower the provided text
        if(self.to_lowercase):
            text = text.lower()

        # perform cleaning while ignoring exclude_cleaning_functions
        if len(self.exclude_cleaning_functions):
            for cleaning_name in dict_cleaning_functions.keys():
                if cleaning_name not in self.exclude_cleaning_functions:
                    text = dict_cleaning_functions[cleaning_name](text)
        # if exclude_cleaning_functions was provided then include-cleaning_functions will be ignoned
        else:
            for cleaning_name in dict_cleaning_functions.keys():
                if cleaning_name in self.include_cleaning_functions:
                    text = dict_cleaning_functions[cleaning_name](text)

        return text



# native packs
import requests
import json
# installed packs
import pandas as pd
from pandas import json_normalize


auth_endpoint = "https://auth.emsicloud.com/connect/token"  # auth endpoint

# replace 'your_client_id' with your client id from your api invite email
client_id = "5f379hywuvh7fvan"
# replace 'your_client_secret' with your client secret from your api invite email
client_secret = "hfCkXQEy"
scope = "emsi_open"  # ok to leave as is, this is the scope we will used

# set credentials and scope
payload = f"client_id={client_id}&client_secret={client_secret}&grant_type=client_credentials&scope={scope}"
# headers for the response
headers = {'content-type': 'application/x-www-form-urlencoded'}
access_token = json.loads((requests.request("POST", auth_endpoint, data=payload, headers=headers)).text)[
    'access_token']  # grabs request's text and loads as JSON, then pulls the access token from that


def fetch_skills_list() -> pd.DataFrame:

    # List of all skills endpoint
    all_skills_endpoint = "https://emsiservices.com/skills/versions/latest/skills"
    # Auth string including access token from above
    auth = f"Authorization: Bearer {access_token}"
    headers = {'authorization': auth}  # headers
    response = requests.request(
        "GET", all_skills_endpoint, headers=headers)  # response
    response = response.json()['data']  # the data

    # all_skills_df = pd.DataFrame(json_normalize(response)); # Where response is a JSON object drilled down to the level of 'data' key
    return response
